require 'core/appkit/lua/class'

OculusVRSystem = Appkit.class(OculusVRSystem)

local Application        = stingray.Application
local Viewport           = stingray.Viewport
local Window             = stingray.Window
local World              = stingray.World
local Unit               = stingray.Unit
local Vector3            = stingray.Vector3
local Quaternion         = stingray.Quaternion
local Matrix4x4          = stingray.Matrix4x4
local Camera             = stingray.Camera
local ShadingEnvironment = stingray.ShadingEnvironment
local Oculus             = stingray.Oculus

OculusVRSystem.generic_tracker_map = {
    HMD = Oculus.LINK_HMD,
    LeftHand  = Oculus.CONTROLLER_LEFT_TOUCH,
    RightHand = Oculus.CONTROLLER_RIGHT_TOUCH
}

OculusVRSystem.generic_button_map = {
	Menu = Oculus.BUTTON_ENTER,
	Trigger = Oculus.BUTTON_LSHOULDER
}

OculusVRSystem.controller_map = {
    Xbox       = Oculus.CONTROLLER_XBOX,
    Remote     = Oculus.CONTROLLER_REMOTE,
    TouchLeft  = Oculus.CONTROLLER_LEFT_TOUCH,
    TouchRight = Oculus.CONTROLLER_RIGHT_TOUCH
}

function OculusVRSystem:init(world)
	if not Oculus then return end

	self._world          = world
	self._enabled        = false
	self._is_initialized = false
	self._controllers 	 = {}
end

function OculusVRSystem:initialize(near, far, hmd_rt, hud_rt)
	if not Oculus then return end
	if not self then return end

    	-- Create main vr camera
	self._vr_main_camera_unit = World.spawn_unit(self._world, "core/units/camera")
	self._vr_main_camera = Unit.camera(self._vr_main_camera_unit, "camera")
	Camera.set_near_range(self._vr_main_camera, near)
	Camera.set_far_range(self._vr_main_camera, far)

	-- Intialize VR system
	self._is_initialized = Oculus.initialize(hmd_rt, self._vr_main_camera, self._world)

	-- Bail if there was an error on initialization
	if not self._is_initialized then
        World.destroy_unit(self._world, self._vr_main_camera_unit)
	    return
	end

    -- Create other vr cameras for blit and hud passes
	self._vr_blit_camera_unit = World.spawn_unit(self._world, "core/units/camera")
	self._vr_hud_camera_unit = World.spawn_unit(self._world, "core/units/camera")
	self._vr_blit_camera = Unit.camera(self._vr_blit_camera_unit, "camera")
	self._vr_hud_camera = Unit.camera(self._vr_hud_camera_unit, "camera")

	-- Create viewports
	self._vr_viewport_main = Application.create_viewport(self._world, "vr_main" )
	self._vr_viewport_blit = Application.create_viewport(self._world, "vr_blit" )
	self._vr_viewport_hud  = Application.create_viewport(self._world, "vr_hud" )

	-- Create VR working window
	local size_x, size_y = Application.render_setting("vr_hmd_resolution")
	local scale          = Application.render_setting("vr_target_scale")
	-- When running in SLI mode, each eye is processed by each GPU and thus only half the size of the full width is needed.
	if Application.render_setting("nv_vr_sli_enabled") then
		size_x = size_x / 2
	end
	self._vr_window      = Window.open{visible = false, width = (size_x * scale), height = (size_y * scale), frameless = true}
	
	-- Setup Oculus HUD layer. Defaults to 0.5m away from HMD
	Oculus.create_layer("hud_layer", hud_rt)
	-- Adding a bit of padding to layer dimensions so that edges are not noticable
	Oculus.set_layer_dimensions("hud_layer", 1.1, 1.2);

	Camera.set_mode(self._vr_blit_camera, Camera.STEREO)
	Camera.set_vertical_fov(self._vr_blit_camera, Camera.vertical_fov(self._vr_blit_camera, 1), 2);

	Window.set_focus()
end

--[[
	Enable VR rendering
]]--
function OculusVRSystem:enable()
	if not Oculus then return end
	if not self or not self._is_initialized then return end

	self._enabled = true
	Oculus.enable(self._enabled);
	Application.set_render_setting("vr_enabled", "true")
end

--[[
	Disable VR rendering
--]]
function OculusVRSystem:disable()
	if not Oculus then return end
	if not self or not self._is_initialized then return end

	self._enabled = false
	Oculus.enable(self._enabled);
	Application.set_render_setting("vr_enabled", "false")
end

--[[
	Returns if VR rendering currently enabled
]]--
function OculusVRSystem:is_enabled()
	if not Oculus then return end
	if not self or not self._is_initialized then return false end

	return self._enabled
end

--[[
	Returns if VR System is initialzed
]]--
function OculusVRSystem:is_initialized()
	if not Oculus then return end
	if not self then return false end

	return self._is_initialized
end


--[[
	Returns VR render buffer size
]]--
function OculusVRSystem:vr_back_buffer_size()
	if not Oculus then return nil end
	if not self or not self._is_initialized then return nil end

	return Application.back_buffer_size(self._eye_window)
end

--[[
	Shutdown VR system
--]]
function OculusVRSystem:shutdown()
	if not Oculus then return end
	if not self or not self._is_initialized then return end

	OculusVRSystem.disable(self)
	Oculus.shutdown()
	
	World.destroy_unit(self._world, self._vr_main_camera_unit)
	World.destroy_unit(self._world, self._vr_blit_camera_unit)
	World.destroy_unit(self._world, self._vr_hud_camera_unit)

	if self._controller1_unit then World.destroy_unit(self._world, self._controller1_unit) end
	if self._controller2_unit then World.destroy_unit(self._world, self._controller2_unit) end

	Application.destroy_viewport(self._world, self._vr_viewport_main)
	Application.destroy_viewport(self._world, self._vr_viewport_blit)
	Application.destroy_viewport(self._world, self._vr_viewport_hud)


	Window.close(self._vr_window)
	self._vr_window = nil

	self._is_initialized = false
end

--[[
	Returns the current Oculus profile values specified in meters such as:
		- profile.player_height
		- profile.eye_height
		- profile.ipd
		- profile.eye_to_neck_x
		- profile.eye_to_neck_y
--]]
function OculusVRSystem:profile()
	if not Oculus then return end
	if not self or not self._is_initialized then return nil end

	return Oculus.profile()
end

--[[
	Returns the current HMD info values such as:
		- hmd_info.product_name
		- hmd_info.manufacturer          
		- hmd_info.resolution_w
		- hmd_info.resolution_h
		- hmd_info.refresh_rate
--]]
function OculusVRSystem:hmd_info()
	if not Oculus then return end
	if not self or not self._is_initialized then return nil end

	return self._hmd_info;
end

--[[
	Recenters the Oculus pose
]]--
function OculusVRSystem:recenter()
	if not Oculus then return end
	if not self or not self._is_initialized then return end

	Oculus.recenter()
end

--[[
	Returns HMD forward vector and local eye poses
]]--
function OculusVRSystem:hmd_local_pose()
	if not Oculus then return nil end
	if not self or not self._is_initialized then return nil end

	head, left, right = Oculus.hmd_local_pose()
	return {
		forward   = Matrix4x4.forward(head),
		left_eye  = left,
		right_eye = right
	}
end

--[[
	Returns HMD forward vector and world eye poses
--]]
function OculusVRSystem:hmd_world_pose()
	if not Oculus then return nil end
	if not self or not self._is_initialized then return nil end

	head, left, right = Oculus.hmd_world_pose()
	return {
		forward   = Matrix4x4.forward(head),
		left_eye  = left,
		right_eye = right
	}
end

function OculusVRSystem:spawn_controllers()
    if not Oculus then return nil end
    if not self or not self._is_initialized then return nil end
    
	-- Create VR controllers
	self._controller1_unit = World.spawn_unit(self._world, "vr_oculus/models/controller/vr_controller_oculus_left")
	Unit.set_data(self._controller1_unit, "tracker_name", "LeftHand")	
	Unit.flow_event(self._controller1_unit, "link_controller")
	self._controllers["LeftHand"] = self._controller1_unit
	
	self._controller2_unit = World.spawn_unit(self._world, "vr_oculus/models/controller/vr_controller_oculus_right")
	Unit.set_data(self._controller2_unit, "tracker_name", "RightHand")		
	Unit.flow_event(self._controller2_unit, "link_controller")
	self._controllers["RightHand"] = self._controller2_unit
end    

function OculusVRSystem:controllers()
	return self._controllers
end

--[[
	Returns the specified controller local pose (either Left or Right Oculus Touch controllers).
--]]
function OculusVRSystem:controller_local_pose(controller_type)
	if not Oculus then return nil end
	if not self or not self._is_initialized then return nil end

	return Oculus.controller_local_pose(controller_type)
end

--[[
	Returns the specified controller world pose (either Left or Right Oculus Touch controllers).
--]]
function OculusVRSystem:controller_world_pose(controller_type)
	if not Oculus then return nil end
	if not self or not self._is_initialized then return nil end

	return Oculus.controller_world_pose(controller_type)
end

--[[
	Returns true/false if specified button is pressed
--]]
function OculusVRSystem:controller_pressed(controller_type, button)
	if not Oculus then return false end
	if not self or not self._is_initialized then return false end
	
	return Oculus.controller_pressed(controller_type, button)
end

--[[
	Returns true/false if specified button is held
--]]
function OculusVRSystem:controller_held(controller_type, button)
	if not Oculus then return false end
	if not self or not self._is_initialized then return false end
	
	return Oculus.controller_held(controller_type, button)
end

--[[
	Returns true/false if specified button is released
--]]
function OculusVRSystem:controller_released(controller_type, button)
	if not Oculus then return false end
	if not self or not self._is_initialized then return false end

	return Oculus.controller_released(controller_type, button)
end

--[[
	Returns true/false if specified button is touched
--]]
function OculusVRSystem:controller_touched(touch_button)
	if not Oculus then return false end
	if not self or not self._is_initialized then return false end

	return Oculus.controller_touched(touch_button)
end

--[[
	Returns true/false if specified button is untouched
--]]
function OculusVRSystem:controller_untouched(touch_button)
	if not Oculus then return false end
	if not self or not self._is_initialized then return false end

    return Oculus.controller_untouched(touch_button)
end

--[[
	Get controller button analog values
--]]
function OculusVRSystem:controller_value(controller_type, button)
	if not Oculus then return nil end
	if not self or not self._is_initialized then return nil end

	local v1, v2 = Oculus.controller_value(controller_type, button)
	return {
	    x = v1,
	    y = v2
	}
end

--[[
	Trigger feedback on the specified controller
--]]
function OculusVRSystem:trigger_controller_feedback(controller_type, frequency, amplitude)
	if not Oculus then return end
	if not self or not self._is_initialized then return end

	Oculus.controller_pulse(controller_type, frequency, amplitude)
end

--[[
	Sets tracking space pose
--]]
function OculusVRSystem:set_tracking_space(translation, rotation, scale)
	if not Oculus then return end
	if not self or not self._is_initialized then return end

	Oculus.set_tracking_space(translation, rotation, scale)
end

--[[
	Render in VR mode
]]--
function OculusVRSystem:render(shading_environment)
	if not Oculus then return end
	if not self or not self._enabled then return end
	
	-- Render HUD
	Application.render_world(self._world, self._vr_hud_camera, self._vr_viewport_hud, shading_environment, self._vr_window)

	-- Render world in instanced stereo
	Application.render_world(self._world, self._vr_main_camera, self._vr_viewport_main, shading_environment, self._vr_window)

	-- Blit result to application window
	Application.render_world(self._world, self._vr_blit_camera, self._vr_viewport_blit, shading_environment)
	return true
end

return OculusVRSystem
