VRToolsFlowCallbacks = VRToolsFlowCallbacks or {}
require 'vr_tools/script/lua/vr_system'
VRSystem = VRSystem or nil
VRApi = VRApi or nil

local Unit          = stingray.Unit
local World         = stingray.World
local Matrix4x4     = stingray.Matrix4x4
local Matrix4x4Box  = stingray.Matrix4x4Box
local Vector3       = stingray.Vector3
local Vector3Box    = stingray.Vector3Box
local Color         = stingray.Color
local SteamVR       = stingray.SteamVR
local LineObject    = stingray.LineObject
local Quaternion    = stingray.Quaternion
local Actor         = stingray.Actor
local Material      = stingray.Material
local Mesh          = stingray.Mesh

local tracking_space_map = {
	["Floor Level"] = stingray.Oculus.TRACKING_ORIGIN_FLOOR_LEVEL,
	["Eye Level"] = stingray.Oculus.TRACKING_ORIGIN_EYELEVEL
}

local function is_world(space)
	return (space == "World")
end

local function get_controller_index(tracker)
	local tracker_id = VRSystem.generic_tracker_map[tracker]
	for i, v in pairs(VRSystem.controller_map) do
		if v == tracker_id then
			return tostring(i)
		end
	end
end

function VRToolsFlowCallbacks.get_controller_index(t)
	if not VRApi then return t end
	t.index = get_controller_index(t.tracker)
	return t
end

function VRToolsFlowCallbacks.vr_system_name(t)
	if not VRSystem then return t end
	t.vrSystem = VRSystem.system_name
	return t
end

function VRToolsFlowCallbacks.hmd_pose(t)
	if not VRApi then return t end

	local head, left, right
	if is_world(t.space) then
		head, left, right = VRApi.hmd_world_pose()
	else
		head, left, right = VRApi.hmd_local_pose()
	end

	t.headPosition     = Matrix4x4.translation(head)
	t.headRotation     = Matrix4x4.rotation(head)
	t.leftEyePosition  = Matrix4x4.translation(left)
	t.leftEyeRotation  = Matrix4x4.rotation(left)
	t.rightEyePosition = Matrix4x4.translation(right)
	t.rightEyeRotation = Matrix4x4.rotation(right)

	return t
end

function VRToolsFlowCallbacks.get_tracking_space(t)
	if not VRApi then return t end
	if not VRSystem then return t end

	local space = VRApi.tracking_space_pose()
	t.position = Matrix4x4.translation(space)
	t.rotation = Matrix4x4.rotation(space)
	t.scale = Matrix4x4.scale(space)
	return t
end

function VRToolsFlowCallbacks.set_tracking_space(t)
	if not VRApi then return t end
	if not VRSystem then return t end

	local space = VRApi.tracking_space_pose()
	local position = t.position or Matrix4x4.translation(space)
	local rotation = t.rotation or Matrix4x4.rotation(space)
	local pose = Matrix4x4.from_quaternion_position(rotation, position)
	local scale = t.scale or Vector3(1, 1, 1)
	Matrix4x4.set_scale(pose, scale)

	Matrix4x4.set_translation(pose, position)
	VRApi.set_tracking_space(position, rotation, scale)
	if VRSystem.system_name == "oculus" then
		local origin = tracking_space_map[t.origin] or stingray.Oculus.TRACKING_ORIGIN_FLOOR_LEVEL
		VRApi.set_tracking_origin(origin)
	end
	return t
end

function VRToolsFlowCallbacks.tracking_space_rectangle(t)
	if not VRApi then return t end
	if not VRSystem then return t end

	if VRSystem.system_name == "oculus" then
		t.corner3 = Vector3(-0.5,0.5,0)
		t.corner4 = Vector3(0.5,0.5,0)
		t.corner1 = Vector3(0.5,-0.5,0)
		t.corner2 = Vector3(-0.5,-0.5,0)
		return t
	end

	if is_world(t.space) then
		t.corner1, t.corner2, t.corner3, t.corner4 = VRApi.tracking_space_world_rectangle()
	else
		t.corner1, t.corner2, t.corner3, t.corner4 = VRApi.tracking_space_local_rectangle()
	end

	return t
end

function VRToolsFlowCallbacks.controller_feedback(t)
	if not VRApi then return t end
	if not VRSystem then return t end

	local controller_index = VRSystem.generic_tracker_map[t.tracker]
	if VRSystem.system_name == "oculus" then
		VRApi.controller_pulse(controller_index, t.frequency or 0.5, t.amplitude or 0.5)
	else
		VRApi.controller_pulse(controller_index, t.seconds or 0.003)
	end
	return t
end


function VRToolsFlowCallbacks.link_node_to_tracker(t)
	if not VRApi then return t end

	local node  = 1
	if t.node_name then
		node = Unit.node(t.unit, t.node_name)
	end

	local world = stingray.Application.flow_callback_context_world()

	local link_pose
	if t.link_to == "HMD" then
		VRApi.link_node_to_tracker(t.unit, node, world, VRApi.LINK_HMD)

		local hmd, _, _ = VRApi.hmd_world_pose()
		link_pose = hmd
	else
		local tracker = VRSystem.generic_tracker_map[t.link_to]

		VRApi.link_node_to_tracker(t.unit, node, world, VRApi.LINK_CONTROLLER, tracker)
		link_pose = VRApi.controller_world_pose(tracker)
	end


	if t.preserve_world_pose then
		-- Calculate position offset
		local inv_link_pose = Matrix4x4.inverse(link_pose)
		local unit_pos = Unit.world_position(t.unit, node)
		local offset_pos = Matrix4x4.transform(inv_link_pose, unit_pos)
		--Calculate rotation offset
		local inv_link_rot = Quaternion.inverse(Matrix4x4.rotation(link_pose))
		local offset_rot = Quaternion.multiply(inv_link_rot, Unit.world_rotation(t.unit, node))
		--Set unit local position to retain offset from tracker
		Unit.teleport_local_position(t.unit, node, offset_pos)
		Unit.teleport_local_rotation(t.unit, node, offset_rot)
	end

end

function VRToolsFlowCallbacks.unlink_node_from_tracker(t)
	if not VRApi then return t end

	local node = 1
	if t.node_name then
		node = Unit.node(t.unit, t.node_name)
	end
	if stingray.Unit.alive(t.unit) then
		VRApi.unlink_node_from_tracker(t.unit, node)
	end
end

function VRToolsFlowCallbacks.controller_pose(t)
	if not VRApi then return t end
	local tracker = VRSystem.generic_tracker_map[t.controller]

	local pose
	if is_world(t.space) then
		pose = VRApi.controller_world_pose(tracker)
	else
		pose = VRApi.controller_local_pose(tracker)
	end

	t.position = Matrix4x4.translation(pose)
	t.rotation = Matrix4x4.rotation(pose)

	return t
end
